//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

//#include "IvoryDraw.h"
//---------------------------------------------------------------------------
static inline TIvorySurface* ValidCtrCheckSurface()
{
	return new TIvorySurface(NULL);
}
//---------------------------------------------------------------------------
__fastcall TIvorySurface::TIvorySurface(Classes::TComponent* AOwner)
 : TComponent(AOwner), FDirectDrawSurface(NULL), FDirectDrawClipper(NULL),
 	FSurfaceType(stUnspecified), FWidth(640), FHeight(480), FTransparent(false),
   FTransparentColor(0), FDrawBackground(false), FMapper(NULL),
   FMapLeft(0), FMapTop(0), FMapColumns(0), FMapRows(0), FMapEnabled(false),
   FMap(NULL), FBackgroundBitmap(NULL), FClearBackground(true),
   FTileBackground(false), FBackgroundColor((TColor)0), FOnPaint(NULL),
   FOnCreate(NULL), FOnDestroy(NULL), FOnGetMap(NULL), FAutoCreate(true)
{
	FCanvas = new TIvoryCanvas(); // create a new canvas
   Canvas->OnChange = OnChange;
}
//---------------------------------------------------------------------------
__fastcall TIvorySurface::~TIvorySurface()
{
	if(DirectDrawSurface)
		TheIvoryDraw->ReleaseSurface(this);
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::BeforeDestruction()
{
	if(OnDestroy) OnDestroy(this);
}
//---------------------------------------------------------------------------
// release the Direct Draw Surface object (clear it from the memory)
void __fastcall TIvorySurface::Release()
{
   /*
   	if(DirectDrawClipper)
      	DirectDrawClipper->Release(); // I'm not really sure if it needs here
   */
//	delete FMapper; // delete the mapper
	ReleaseCanvas(); // release the device context grabbed from the surface
//	delete FCanvas; // delete the canvas
	if(DirectDrawSurface)
   {
   	if(SurfaceType != stBack) // if it's not a back surface
   		DirectDrawSurface->Release(); // release the Direct Draw Surface
      FDirectDrawSurface = NULL;
   }
}
//---------------------------------------------------------------------------
// get a device context from the surface
void __fastcall TIvorySurface::GrabCanvas()
{
	if(DirectDrawSurface)
   {
		HDC hdc;
   	HRESULT ddrval;
		ddrval = DirectDrawSurface->GetDC(&hdc); // get the context
		if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvorySurface::GrabCanvas_GetDC");
		Canvas->SetHandle(hdc); // pass the context to the canvas object
   }
}
//---------------------------------------------------------------------------
// release the device context grabbed from the surface
void __fastcall TIvorySurface::ReleaseCanvas()
{
	if(Canvas->Handle)
   {
   	DirectDrawSurface->ReleaseDC(Canvas->Handle); // release the context
		Canvas->SetHandle(NULL); // pass NULL to the canvas object
   }
}
//---------------------------------------------------------------------------
// create primary surface
void __fastcall TIvorySurface::CreatePrimary()
{
	if(DirectDrawSurface == NULL)
   {
		HRESULT ddrval;
		FSurfaceType = stPrimary; // gonna be a primary surface
      DDSURFACEDESC ddsd;
      ZeroMemory(&ddsd, sizeof(ddsd));
      ddsd.dwSize = sizeof(ddsd);
      if(TheIvoryDraw->ExclusiveMode) // if we're in exclusive mode
      {
	      ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
   	   ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
	      ddsd.dwBackBufferCount = 1;
      }
      else // if we're not in exclusive mode
      {
	      ddsd.dwFlags = DDSD_CAPS;
   	   ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
      }
      ddrval = TheIvoryDraw->DirectDraw->CreateSurface(&ddsd, &FDirectDrawSurface, NULL);
  	   if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvorySurface::CreatePrimary_CreateSurface");
      FWidth = 640; FHeight = 480;
	   if(OnCreate) OnCreate(this);
   }
}
//---------------------------------------------------------------------------
// create back surface by getting the back surface from the primary surface
void __fastcall TIvorySurface::CreateBack(TIvorySurface* primary)
{
	if(DirectDrawSurface == NULL)
   {
		HRESULT ddrval;
   	if(TheIvoryDraw->ExclusiveMode) // if we're in exclusive mode
      {
			DDSCAPS ddscaps;
			FSurfaceType = stBack; // gonna be a back surface
			ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
			ddrval = primary->DirectDrawSurface->GetAttachedSurface(&ddscaps, &FDirectDrawSurface);
	  	   if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvorySurface::CreateBack_CreateSurface#1");
      }
      else // if we're not in exclusive mode
      {
			HRESULT ddrval;
   		DDSURFACEDESC ddsd;

			FSurfaceType = stOffScreen; // gonna be an off-screen surface
			ZeroMemory(&ddsd, sizeof(ddsd));
	      ddsd.dwSize = sizeof(ddsd);
   	   ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
      	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	      ddsd.dwHeight = 480;
   	   ddsd.dwWidth = 640;
      	ddrval = TheIvoryDraw->DirectDraw->CreateSurface(&ddsd, &FDirectDrawSurface, NULL);
	  	   if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvorySurface::CreateBack_CreateSurface#2");

         ddrval = TheIvoryDraw->DirectDraw->CreateClipper(0, &FDirectDrawClipper, NULL);
	  	   if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvorySurface::CreateBack_CreateClipper");
         ddrval = DirectDrawClipper->SetHWnd(0, TheIvoryDraw->Handle);
	  	   if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvorySurface::CreateBack_SetHWnd");
         ddrval = primary->DirectDrawSurface->SetClipper(DirectDrawClipper);
	  	   if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvorySurface::CreateBack_SetClipper");
      }
      FWidth = 640; FHeight = 480;
      CreateMapperSurface();
	   if(OnCreate) OnCreate(this);
//      Paint();
   }
}
//---------------------------------------------------------------------------
// create off-screen surface; FWidth and FHeight must be set before calling it
void __fastcall TIvorySurface::_CreateOffScreen()
{
	if(TheIvoryDraw == NULL || TheIvoryDraw->DirectDraw == NULL) throw EIvoryError(iecCreateSurfaceError, "TIvorySurface::_CreateOffScreen_DirectDraw");
   if(DirectDrawSurface)
   {
   	if(SurfaceType != stOffScreen) return; // primary and back surfaces cannot be recreated
      BeforeDestruction();
      Release();
   }

	FSurfaceType = stOffScreen; // gonna be an off-screen surface

	HRESULT ddrval;
  	DDSURFACEDESC ddsd;

	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwHeight = Height;
	ddsd.dwWidth = Width;
	ddrval = TheIvoryDraw->DirectDraw->CreateSurface(&ddsd, &FDirectDrawSurface, NULL);
	if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvorySurface::CreateOffScreen_CreateSurface");
   TheIvoryDraw->AddSurface(this);
	if(Transparent) // if transparent, set the transparent color
	{
		DDCOLORKEY ddck;
		ddck.dwColorSpaceLowValue = TransparentColor;
		ddck.dwColorSpaceHighValue = TransparentColor;
		DirectDrawSurface->SetColorKey(DDCKEY_SRCBLT, &ddck);
	}
   CreateMapperSurface();

	if(OnCreate) OnCreate(this);
	Paint(); // paint the surface
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::CreateOffScreen()
{
	if(DirectDrawSurface == NULL && Width > 0 && Height > 0) _CreateOffScreen();
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::RecreateOffScreen(int width, int height)
{
	FWidth = width;
   FHeight = height;
   _CreateOffScreen();
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::RecreateOffScreen(int width, int height, bool transparent, int color)
{
	FWidth = width;
   FHeight = height;
   FTransparent = transparent;
   FTransparentColor = color;
   _CreateOffScreen();
}
//---------------------------------------------------------------------------
void __fastcall TIvorySurface::SetAutoCreate(bool ac)
{
	if(DirectDrawSurface == NULL) // AutoCreate can only be set before creating an off-screen surface
   	FAutoCreate = ac;
}
//---------------------------------------------------------------------------
// sets the width of the surface
void __fastcall TIvorySurface::SetWidth(int width)
{
	if(DirectDrawSurface == NULL) // the width can only be set before creating an off-screen surface
		FWidth = width;
}
//---------------------------------------------------------------------------
// sets the height of the surface
void __fastcall TIvorySurface::SetHeight(int height)
{
	if(DirectDrawSurface == NULL) // the height can only be set before creating an off-screen surface
   	FHeight = height;
}
//---------------------------------------------------------------------------
// sets the transparency of the surface
void __fastcall TIvorySurface::SetTransparent(bool transparent)
{
	if(DirectDrawSurface == NULL) // the transparency can only be set before creating an off-screen surface
   	FTransparent = transparent;
}
//---------------------------------------------------------------------------
// sets the transparent color of the surface
void __fastcall TIvorySurface::SetTransparentColor(int tcolor)
{
	if(DirectDrawSurface == NULL) // the transparent color can only be set before creating an off-screen surface
   	FTransparentColor = tcolor;
}
//---------------------------------------------------------------------------
// paint the surface
void __fastcall TIvorySurface::Paint()
{
	if(TheIvoryDraw == NULL || TheIvoryDraw->DirectDraw == NULL) throw EIvoryError(iecPaintSurfaceError, "TIvorySurface::Paint_DirectDraw");
	if(DirectDrawSurface == NULL) throw EIvoryError(iecPaintSurfaceError, "TIvorySurface::Paint_DirectDrawSurface");
   if(BeforePaint) BeforePaint(this);
   if(OnPaint || DrawBackground || ClearBackground)
   {
		GrabCanvas(); // get a device context
		if(DrawBackground && BackgroundBitmap) // if there is a background bitmap
	   {
   		if(TileBackground)
      	{
	      	TRect r; r.Left = 0; r.Top = 0; r.Right = Width; r.Bottom = Height;
			   Canvas->TileDraw(r, BackgroundBitmap); // draw the background bitmap tiled
      	}
	      else
   	   {
      		TRect r; r.Left = 0; r.Top = 0; r.Right = Width; r.Bottom = Height;
		   	Canvas->StretchDraw(r, BackgroundBitmap); // draw the background bitmap stretched
	      }
   	}
	   else if(ClearBackground)
   	{
   		Canvas->Pen->Color = FBackgroundColor;
	   	Canvas->Brush->Color = FBackgroundColor;
   	   Canvas->Rectangle(0, 0, Width, Height);
	   }
      /*
      for(int x = 0; x < 16; x++)
      for(int y = 0; y < 16; y++)
      {
      	HBRUSH OldBrush, Brush;
      	Brush = CreateSolidBrush(x+y*16);
         OldBrush = SelectObject(Canvas->Handle, Brush);
         Rectangle(Canvas->Handle, x*10,y*10,x*10+9,y*10+9);
         SelectObject(Canvas->Handle, OldBrush);
         DeleteObject(Brush);
      }//onyl for debug
      */
   	if(OnPaint) // if there is an OnPaint event handler
	   	OnPaint(this); // call the OnPaint event
   	ReleaseCanvas(); // release the device context
   }
   if(AfterPaint) AfterPaint(this);
   if(MapEnabled && Mapper)
	  	DrawMap();
   if(AfterDrawMap) AfterDrawMap(this);
}
//---------------------------------------------------------------------------
// Copy another surface
void __fastcall TIvorySurface::CopySurface(int origx, int origy, TIvorySurface* surface)
{
	CopySurface(origx, origy, surface, Rect(0, 0, surface->Width, surface->Height));
}
//---------------------------------------------------------------------------
// Copy a rectangle from another surface
void __fastcall TIvorySurface::CopySurface(int origx, int origy, TIvorySurface* surface, const TRect rect)
{
	if(surface == NULL || surface->DirectDrawSurface == NULL) throw EIvoryError(iecSourceSurfaceError, "TIvorySurface::CopySurface_DirectDrawSurface");
   int options;
   if(surface->Transparent) options = DDBLTFAST_SRCCOLORKEY;
   else options = DDBLTFAST_NOCOLORKEY;
	if(SurfaceType == stBack) options |= DDBLTFAST_WAIT;
	HRESULT ddrval;
   ddrval = DirectDrawSurface->BltFast(origx, origy, surface->DirectDrawSurface, (RECT*)&rect, options);
   if(ddrval != DD_OK) throw EIvoryError(ddrval, "TIvorySurface::CopySurface_BltFast");
   if(OnChange) OnChange(this);
}
//---------------------------------------------------------------------------
// if the mapper surface is not yet created, this function creates it
void __fastcall TIvorySurface::CreateMapperSurface()
{
	if(Mapper && Mapper->Surface->BackgroundBitmap->Handle && Mapper->Surface->DirectDrawSurface == NULL)
   // If there's a mapper, and its surface is not yet created, we need to create it
	{
		Mapper->Surface->RecreateOffScreen(Mapper->Surface->BackgroundBitmap->Width, Mapper->Surface->BackgroundBitmap->Height);
	}
}
//---------------------------------------------------------------------------
// draw a map on the surface. A mapper must be set and the OnGetMap event must be written
void __fastcall TIvorySurface::DrawMap()
{
	if(/*MapEnabled && Mapper &&*/ MapRows && MapColumns && (OnGetMap || Map))
   {
   	int maxindex = min(Mapper->TileColumns * Mapper->TileRows, Mapper->Tiles);
   	int index = 0;
      int x, y;
      int row, col;
      if(Map) // if Map is specified, use it
      {
      	int* map = Map;
			for(row = 0, y = MapTop; row < MapRows; row++, y += Mapper->TileHeight)
   	   {
      		for(col = 0, x = MapLeft; col < MapColumns; col++, x += Mapper->TileWidth)
         	{
            	index = *map++;
	            if(index >= 0 && index < maxindex)
						CopySurface(x, y, Mapper->Surface, Mapper->TileRect[index]);
      	   }
	      }
      }
      else // if Map is NULL, use OnGetMap
      {
			for(row = 0, y = MapTop; row < MapRows; row++, y += Mapper->TileHeight)
   	   {
      		for(col = 0, x = MapLeft; col < MapColumns; col++, x += Mapper->TileWidth)
         	{
            	index = -1;
         		OnGetMap(this, row, col, index);
	            if(index >= 0 && index < maxindex)
						CopySurface(x, y, Mapper->Surface, Mapper->TileRect[index]);
      	   }
	      }
      }
   }
}
//---------------------------------------------------------------------------

